from io import BytesIO
from pathlib import PurePosixPath

import lazy_object_proxy
from logbook import Logger
from PIL import Image
from PIL import UnidentifiedImageError
from google.cloud import storage

from .common import GCFContext, ImgSize, Thumbnail
from .conf import get_monitored_paths, get_thumbnail_sizes


SUBFOLDER = 'thumbnails'
logger = Logger(__name__)
# Laziyly instantiate Google Cloud client, so that it won't break unittest
store = lazy_object_proxy.Proxy(storage.Client)


def build_thumbnail_path(original: PurePosixPath, size: ImgSize) -> PurePosixPath:
    '''
    Build file path for thumbnail from original file.

    Example: abc/photo.jpg -> abc/photo_512x512.jpg
    The "orignal" argument has PurePosixPath type because the path is in Google Cloud Storage context,
    not neccessary be real filesytem path.
    '''
    ext = original.suffix
    folder = original.parent
    return folder / SUBFOLDER / f'{original.stem}_{size}{ext}'


def upload(bucket: storage.Bucket, thumb: Thumbnail):
    blob = bucket.blob(thumb.path)
    blob.upload_from_string(thumb.content, thumb.mimetype)
    logger.info('Uploaded {}.', thumb.path)


def create_thumbnail(orig: Image.Image, size: ImgSize, orpath: PurePosixPath) -> Thumbnail:
    img = orig.copy()
    mimetype = orig.get_format_mimetype()
    img.thumbnail(size)
    out = BytesIO()
    img.save(out, orig.format)
    out.seek(0)
    thumbpath = build_thumbnail_path(orpath, size)
    logger.debug('Thumbnail path: {}', thumbpath)
    return Thumbnail(out.getvalue(), thumbpath, size, mimetype)


def generate_gs_thumbnail(data: dict, context: GCFContext):
    '''Background Cloud Function to be triggered by Cloud Storage'''
    event_type = context.event_type
    if event_type != 'google.storage.object.finalize' and event_type != 'google.storage.object.delete':
        return
    if event_type == 'google.storage.object.delete':
        return
    filepath = data['name']   # type: str
    if not any(filepath.startswith(p) for p in get_monitored_paths()):
        logger.error('File {} is not watched. Ignore.', filepath)
        return
    content_type = data['contentType']  # type: str
    if not content_type.startswith('image/'):
        logger.info('The file {} is not an image (content type {}). Ignore.', filepath, content_type)
        return
    filepath = PurePosixPath(filepath)
    folder_name = filepath.parent.name
    if folder_name == SUBFOLDER:
        logger.info('The file {} is already a thumbnail. Ignore.', filepath)
        return
    bucket = store.get_bucket(data['bucket'])
    blob = bucket.get_blob(str(filepath))
    filecontent = blob.download_as_string()
    try:
        orig = Image.open(BytesIO(filecontent))    # type: Image.Image
    except UnidentifiedImageError:
        logger.error('This image {} is not supported by Pillow.', filepath)
        return
    for size in get_thumbnail_sizes():   # type: ImgSize
        thumb = create_thumbnail(orig, size, filepath)
        upload(bucket, thumb)



import os
from typing import Tuple

from dotenv import load_dotenv, find_dotenv

from .common import ImgSize


load_dotenv(find_dotenv(usecwd=True))


def get_thumbnail_sizes() -> Tuple[ImgSize, ...]:
    var = os.getenv('THUMB_SIZES')
    if not var:
        return ()
    sizes = []
    for p in var.split(','):
        spec = p.strip()
        try:
            w, h = spec.split('x')
        except ValueError:
            continue
        try:
            sizes.append(ImgSize(int(w), int(h)))
        except ValueError:
            continue
    return tuple(sizes)


def get_monitored_paths() -> Tuple[str, ...]:
    var = os.getenv('MONITORED_PATHS')
    if not var:
        return ()
    paths = []
    for p in var.split(':'):
        if p:
            paths.append(p)
    return tuple(paths)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['thunagen']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=7.0.0,<8.0.0',
 'google-cloud-storage>=1.24.1,<2.0.0',
 'lazy_object_proxy>=1.4.3,<2.0.0',
 'logbook>=1.5.3,<2.0.0',
 'python-dotenv>=0.10.3,<0.11.0']

setup_kwargs = {
    'name': 'thunagen',
    'version': '0.1.1',
    'description': 'Google Cloud function to generate thumbnail for images in Google Storage.',
    'long_description': '========\nThunagen\n========\n\n\nGoogle Cloud function to generate thumbnail for images in Google Storage.\n\nConvention\n----------\n\nThe thumbnails are placed in a folder "thumbnails" at the same place as original file.\n\nThe thumbnail size is appended to filename, right before the extention part. For example:\n\n\n.. code-block::\n\n    bucket\n    └── folder\n        ├── photo.jpg\n        └── thumbnails\n            ├── photo_128x128.jpg\n            └── photo_512x512.jpg\n\nThe function expect these environment variables to be set:\n\n- ``THUMB_SIZES``: Size of thumbnail to be generated. Example: ``512x512,128x128``.\n\n- ``MONITORED_PATHS``: Folders (and theirs children) where the function will process the uploaded images. Muliple paths are separated by ":", like ``user-docs:user-profiles``.\n\nThe variables can be passed via *.env* file in the working directory.\n\nInclude to your project\n-----------------------\n\nThunagen is provided without a *main.py* file, for you to easier incorporate to your project, where you may have your own way to configure deployment environment (different bucket for "staging" and "production", for example).\n\nTo include Thunagen, from your *main.py*, do:\n\n.. code-block:: py\n\n    from thunagen.functions import generate_gs_thumbnail\n',
    'author': 'Nguyễn Hồng Quân',
    'author_email': 'ng.hong.quan@gmail.com',
    'maintainer': 'Nguyễn Hồng Quân',
    'maintainer_email': 'ng.hong.quan@gmail.com',
    'url': 'https://github.com/sunshine-tech/thunagen.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
